from typing import List

import pytest

from barril.units import Scalar
from barril.units.unit_database import UnitDatabase


class _LightweightQuantity:
    """
    A lightweight representation of a quantity. This class has the following purposes:

    1) To be a "model" implementation (or goal) for the Quantity. The current
       barril.units.Quantity implementation has some advanced features (used in
       very few places) that make it heavy and slow.

    2) To be checked, in terms of performance, against the full Quantity and against
        a tuple representation.

    .. see:: barril.units.Quantity

    This should be as lightweight as possible, storing only strings, and just the information
    needed: the unit (such as 'kg/m3') and the category (such as 'density').

    The LightweightQuantity also assumes that the used UnitDatabase is the singleton one.
    """

    __slots__ = ["_unit", "_category"]

    def __init__(self, unit, category):
        """
        :param str unit:
            The unit name. It must be valid within that category.

        :param str category:
            A valid category.
        """
        self._unit = unit
        self._category = category

    def GetCategory(self):
        return self._category

    def GetUnit(self):
        return self._unit

    def GetQuantityType(self):
        category_info = self.GetUnitDatabase().GetCategoryInfo(self.GetCategory())
        return category_info.quantity_type

    def GetQuantity(self):
        return self

    def GetUnitDatabase(self):
        return UnitDatabase.GetSingleton()


class _LightweightScalar(tuple):  # Could derive from _LightweightQuantity, but this way is faster
    """
    This is a lightweight version of a Scalar object. Should be as lightweight as a namedtuple,
    without events or any other kind of overhead. This class should be convertible to a full
    Scalar.

    Assumptions:
        - Uses the Singleton unit database
        - The category and the unit must be valid (will not be checked at initialization)
    """

    __slots__: List[str] = []

    def __new__(cls, value, unit, category):
        return tuple.__new__(cls, (value, unit, category))

    def GetValue(self, unit=None):
        assert (
            unit is None or unit == self[1]
        ), "LightweightScalar does not support unit conversion."
        return self[0]

    def GetValueAndUnit(self):
        return (self[0], self[1])

    def IsValid(self):
        return self.CheckValue(self[0])  # type:ignore[attr-defined]

    def GetCategory(self):
        return self[2]

    def GetUnit(self):
        return self[1]

    def GetQuantityType(self):
        category_info = self.GetUnitDatabase().GetCategoryInfo(self[2])
        return category_info.quantity_type

    def GetQuantity(self):
        return self

    def GetUnitDatabase(self):
        return UnitDatabase.GetSingleton()

    def __iter__(self):
        """
        Allows unpacking this object into a (value, unit) tuple.
        """
        return iter((self[0], self[1]))


def testCreation() -> None:
    x = _LightweightScalar(value=1.0, unit="kg/m3", category="density")

    assert x.GetValue() == 1.0

    assert x.GetUnit() == "kg/m3"
    assert x.GetCategory() == "density"
    assert x.GetQuantityType() == "density"

    assert x.GetQuantity().GetUnit() == "kg/m3"
    assert x.GetQuantity().GetCategory() == "density"
    assert x.GetQuantity().GetQuantityType() == "density"


def testConversionToFullScalar() -> None:
    y = _LightweightScalar(value=1.0, unit="kg/m3", category="density")
    x = Scalar(value=y.GetValue(), unit=y.GetUnit(), category=y.GetCategory())

    assert x.GetValue() == 1.0

    assert x.GetUnit() == "kg/m3"
    assert x.GetCategory() == "density"
    assert x.GetQuantityType() == "density"

    assert x.GetQuantity().GetUnit() == "kg/m3"
    assert x.GetQuantity().GetCategory() == "density"
    assert x.GetQuantity().GetQuantityType() == "density"

    # Full Scalar supports conversion
    assert x.GetValue(unit="g/cm3") == 0.001


def testLazyChecking() -> None:
    from barril.units.unit_database import InvalidQuantityTypeError

    # Category 'meta density' does not exist
    x = _LightweightScalar(value=0.001, unit="kg/m3", category="meta density")

    # But the creation of _LightweightScalar does not check this...
    assert x.GetValue() == 0.001

    assert x.GetUnit() == "kg/m3"
    assert x.GetCategory() == "meta density"

    # We will get the message error only when trying to get the related quantity type:
    with pytest.raises(InvalidQuantityTypeError):
        x.GetQuantityType()


def testGettingValueWithDifferentUnit() -> None:
    x = _LightweightScalar(value=0.001, unit="g/cm3", category="concentration")

    # But the creation of _LightweightScalar does not check this...

    assert x.GetValue() == 0.001
    assert x.GetValue("g/cm3") == 0.001
    assert x.GetValueAndUnit() == (0.001, "g/cm3")

    assert x.GetUnit() == "g/cm3"
    assert x.GetCategory() == "concentration"
    assert x.GetQuantityType() == "density"

    # We will get an assertion error when trying to get the value with a different unit
    with pytest.raises(AssertionError):
        x.GetValue("kg/cm3")


def testUnpackingToValueAndUnit() -> None:
    x = _LightweightScalar(value=0.001, unit="g/cm3", category="concentration")

    value, unit = x

    assert value == 0.001
    assert unit == "g/cm3"


def testQuantityCreationAndInterface() -> None:
    x = _LightweightQuantity(unit="kg/m3", category="concentration")

    assert x.GetUnit() == "kg/m3"
    assert x.GetCategory() == "concentration"
    assert x.GetQuantityType() == "density"

    assert x.GetQuantity().GetUnit() == "kg/m3"
    assert x.GetQuantity().GetCategory() == "concentration"
    assert x.GetQuantity().GetQuantityType() == "density"


def testQuantityLazyChecking() -> None:
    from barril.units.unit_database import InvalidQuantityTypeError

    # Category 'meta density' does not exist
    x = _LightweightQuantity(unit="kg/m3", category="meta density")

    # But the creation of _LightweightQuantity does not check this...

    assert x.GetUnit() == "kg/m3"
    assert x.GetCategory() == "meta density"

    # We will get the message error only when trying to get the related quantity type:
    with pytest.raises(InvalidQuantityTypeError):
        x.GetQuantityType()



import pytest
from pytest import approx

from barril import units
from barril.units import UnitsError
from barril.units.unit_database import UnitDatabase


def testPoscWithoutFillCategories(unit_database_posc_len_no_category) -> None:
    unit_database = unit_database_posc_len_no_category

    unit_database.AddCategory("my_len", "length")

    # this category does not exist because we didn't fill the categories
    with pytest.raises(UnitsError):
        units.Scalar("length", 100, "km")

    u1 = units.Scalar("my_len", 100, "km")
    u2 = units.Scalar("my_len", 100000, "m")
    assert approx(abs(u1.value - u2.GetValue("km")), 7) == 0


def testMegagramPerCubicMeterToKilogramPerCubicMeter() -> None:
    unit_database = UnitDatabase.CreateDefaultSingleton()
    expected = 1000
    obtained = unit_database.Convert("density", "Mg/m3", "kg/m3", 1)
    assert obtained == expected


@pytest.mark.parametrize(
    "unit, manometric_unit, conversion",
    [
        ("Pa", "Pa(g)", -101225.0),
        ("bar", "bar(g)", 98.9867),
        ("kPa", "kPa(g)", -1.325),
        ("kgf/cm2", "kgf/cm2(g)", 98.9667),
    ],
)
def testManometricPressure(unit: str, manometric_unit: str, conversion: float) -> None:
    scalar = units.Scalar("pressure", 100, unit)
    assert scalar.GetValue(manometric_unit) == approx(conversion)


def testDivisionError() -> None:
    unit_database = UnitDatabase.CreateDefaultSingleton()
    for category, category_info in unit_database.categories_to_quantity_types.items():
        base_unit = category_info.default_unit
        assert base_unit is not None
        for unit in unit_database.GetValidUnits(category):
            for i in [-1, 0, 1]:
                try:
                    unit_database.Convert(category, base_unit, unit, i)
                except Exception as e:
                    raise TypeError(f"Error converting: from: {base_unit} to: {unit}") from e

                try:
                    unit_database.Convert(category, unit, base_unit, i)
                except Exception as e:
                    raise TypeError(f"Error converting: from: {unit} to: {base_unit}") from e


def testScfPerBblToM3ToM3() -> None:
    unit_database = UnitDatabase.CreateDefaultSingleton()
    expected = 0.17776487178535644
    obtained = unit_database.Convert("standard volume per volume", "scf/bbl", "scm(15C)/m3", 1.0)
    assert approx(abs(obtained - expected), 7) == 0


def testAllValidTemperatureUnitsHaveSameDefaultCategory() -> None:
    """'temperature' units should have 'temperature' as defined category (#45)"""
    unit_database = UnitDatabase.CreateDefaultSingleton()

    expected_category = "temperature"
    valid_temperature_units = unit_database.GetValidUnits(expected_category)

    for unit in valid_temperature_units:
        T = units.Scalar(1.0, unit)
        assert T.category == expected_category



import pytest

from barril.units import ObtainQuantity
from barril.units import UnitDatabase


def testReadOnlyQuantity(unit_database_empty) -> None:
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddCategory("length", "length")

    read_only_quantity = ObtainQuantity("m", "length")
    with pytest.raises(AttributeError):
        read_only_quantity.SetUnit("cm")  # type:ignore[attr-defined]

    # When creating a copy of a read only quantity we'll make it not read only anymore!
    copy = read_only_quantity.MakeCopy()
    assert copy.GetUnitDatabase() is unit_database


def testMultiplyReadOnlyQuantity() -> None:
    """
    Raising an Error when Geometry quantity is read only (...)

    This test simulates the conditions of the bug report. This bug has been fixed before I try
    to reproduce here.
    """
    unit_database = UnitDatabase.GetSingleton()

    meters = ObtainQuantity("m", "length")

    r_quantity, _value = unit_database.Multiply(meters, meters, 1.0, 1.0)
    assert r_quantity.GetUnit() == "m2"



from collections import OrderedDict

import pytest

from barril.units import InvalidOperationError
from barril.units import InvalidUnitError
from barril.units import ObtainQuantity
from barril.units import Quantity


def testDerivedQuantities(unit_database_len_time) -> None:
    # define a simple quantity
    ObtainQuantity(unit="s", category="Time")  # see if it works
    ObtainQuantity(unit="m", category="Table size")  # see if it works
    q3 = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 2]), ("Time", ["s", -1])]))
    q4 = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 2]), ("Time", ["s", -2])]))
    q5 = Quantity.CreateDerived(
        OrderedDict([("Table size", ["m", 1]), ("City size", ["m", 1]), ("Time", ["s", -2])])
    )
    q6 = Quantity.CreateDerived(OrderedDict([("Time", ["s", -2])]))
    q7 = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1]), ("Time", ["s", 2])]))

    with pytest.raises(InvalidUnitError):
        Quantity.CreateDerived(
            OrderedDict(
                [("Table size", ["invalid", 1]), ("City size", ["m", 1]), ("Time", ["s", -2])]
            )
        )

    assert "(Table size) ** 2 / Time" == q3.GetCategory()
    assert "m2/s" == q3.GetUnit()

    assert "(Table size) ** 2 / (Time) ** 2" == q4.GetCategory()
    assert "m2/s2" == q4.GetUnit()
    assert "1/s2" == q6.GetUnit()
    assert "m.s2" == q7.GetUnit()

    assert (("m", 2), ("s", -2)) == q4.GetComposingUnits()
    assert (("m", 1), ("m", 1), ("s", -2)) == q5.GetComposingUnits()


def testConvertionWithDerivedUnits(unit_database_len_time) -> None:
    empty = Quantity.CreateDerived(OrderedDict())
    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    m_city = Quantity.CreateDerived(OrderedDict([("City size", ["m", 1])]))
    cm = Quantity.CreateDerived(OrderedDict([("Table size", ["cm", 1])]))
    km_city = Quantity.CreateDerived(OrderedDict([("City size", ["km", 1])]))
    m2 = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 2])]))
    s = Quantity.CreateDerived(OrderedDict([("Time", ["s", -1])]))
    m2s = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 2]), ("Time", ["s", -1])]))
    cat_mix_m2 = Quantity.CreateDerived(
        OrderedDict([("Table size", ["m", 1]), ("City size", ["m", 1])])
    )

    unit_database = unit_database_len_time
    # multiplication
    assert (m2, 2) == unit_database.Multiply(m, m, 1, 2)
    assert (m2s, 2) == unit_database.Multiply(m2, s, 1, 2)
    assert (m2s, 2) == unit_database.Multiply(m2, s, 1, 2)
    assert (m2, 1) == unit_database.Multiply(m, cm, 1, 100)
    assert (cat_mix_m2, 1) == unit_database.Multiply(m, m_city, 1, 1)
    assert (cat_mix_m2, 1) == unit_database.Multiply(m, km_city, 1, 0.001)

    # division
    assert (m, 0.001) == unit_database.Divide(cat_mix_m2, km_city, 1, 1)
    # check division with cancelling units (and different categories)
    assert (empty, 1) == unit_database.Divide(m, m_city, 1, 1)

    # floor division
    assert (m, 3.0) == unit_database.FloorDivide(cat_mix_m2, km_city, 3.5, 0.001)

    # sum
    assert (m, 1 + 0.01) == unit_database.Sum(m, cm, 1, 1)
    assert (m, 2) == unit_database.Sum(m, m_city, 1, 1)

    # subtraction
    # check with operation so that we have an exact match (without need for almost equals,
    # as that should be the same exact operation done later)
    assert (m, 1 - 0.01) == unit_database.Subtract(m, cm, 1, 1)
    with pytest.raises(InvalidOperationError):
        unit_database.Subtract(m, s, 1, 1)


def testDeepcopy(unit_database_len_time) -> None:
    import copy

    # Note: the part below is flaky for a test because it relies on getting a refcount to None
    # which could change if there are other threads running. The code is kept just as a
    # reference in case we have to debug such a situation again (there was a bug in odict where
    # it decreased references to None when it shouldn't and it crashed the program later on).

    # m = OrderedDict([('Table size', ['m', 1])])
    # import gc
    # import sys
    # my_none = None
    # for i in range(100):
    #     gc.collect()
    #
    # Note:
    # for i in range(200):
    #     gc.collect()
    #     if i < 100:
    #         prev = sys.getrefcount(my_none)
    #     else:
    #         curr = sys.getrefcount(my_none)
    #         self.assert_(curr >= prev,
    #             'The ref count cannot get lower for None (previous:%s current:%s).' % (prev, curr))
    #
    #     # Notice that print sys.getrefcount(None) is always decrementing (this is the error)
    #     m = copy.deepcopy(m)

    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    m0 = copy.deepcopy(m)
    assert m is m0  # Check if our cache is working.


def testReadOnlyOperation(unit_database_len_time) -> None:
    unit_database = unit_database_len_time
    m_ro = ObtainQuantity("m", "Table size")
    m_rw = ObtainQuantity("m", "Table size")

    m2 = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 2])]))

    # multiplication
    assert (m2, 2) == unit_database.Multiply(m_rw, m_rw, 1, 2)
    assert (m2, 2) == unit_database.Multiply(m_ro, m_rw, 1, 2)
    assert (m2, 2) == unit_database.Multiply(m_ro, m_ro, 1, 2)
    assert (m2, 2) == unit_database.Multiply(m_rw, m_ro, 1, 2)

    quantity, _ = unit_database.Multiply(m_ro, m_ro, 1, 2)
    assert isinstance(quantity, Quantity)

    quantity, _ = unit_database.Multiply(m_rw, m_rw, 1, 2)
    assert isinstance(quantity, Quantity)
    quantity, _ = unit_database.Multiply(m_rw, m_ro, 1, 2)
    assert isinstance(quantity, Quantity)
    quantity, _ = unit_database.Multiply(m_ro, m_rw, 1, 2)
    assert isinstance(quantity, Quantity)



import operator

import pytest

from barril import units
from barril.units import ObtainQuantity
from barril.units.unit_database import UnitDatabase


def testOperation() -> None:
    unit_database = UnitDatabase.GetSingleton()
    unit_database.CheckDefaultUnitDatabase()
    q = ObtainQuantity(unit="m", category="length")

    SCALAR_OPERATION = [operator.add, operator.sub, operator.truediv, operator.mul, operator.mod]
    SCALAR_TYPES = (int, float)

    for operation in SCALAR_OPERATION:
        for cast_type in SCALAR_TYPES:
            # Left
            q2 = operation(cast_type(2), q)
            assert q2.GetQuantityType() == "length"
            assert q2.GetUnit() == "m"

            # Right
            q2 = operation(q, cast_type(2))
            assert q2.GetQuantityType() == "length"
            assert q2.GetUnit() == "m"

    q2 = abs(q)
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q + q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q - q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * int(2)
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * 2.0
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q * q
    assert q2.GetQuantityType() == "(length) ** 2"
    assert q2.GetUnit() == "m2"

    q2 = q * q * q
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    # Check pow
    q2 = q**1
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = q**2
    assert q2.GetQuantityType() == "(length) ** 2"
    assert q2.GetUnit() == "m2"

    q2 = q**3
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    q2 = pow(q, 3)
    assert q2.GetQuantityType() == "(length) ** 3"
    assert q2.GetUnit() == "m3"

    q2 = (q * q) / q
    assert q2.GetQuantityType() == "length"
    assert q2.GetUnit() == "m"

    q2 = (q * q) / ObtainQuantity(unit="d", category="time")
    assert q2.GetQuantityType() == "(length) ** 2 / time"
    assert q2.GetUnit() == "m2/d"

    with pytest.raises(TypeError):
        operator.mul(q, "s")


def MakeArray(values):
    """Creates a standard array used by testScalarOperations"""
    return units.Array("length", values, "m")


def MakeFixedArray(values):
    """Creates a fixed array with 3 dimensions used by testScalarOperations"""
    return units.FixedArray(3, "length", values, "m")


@pytest.mark.parametrize(
    "left, op, right, expected",
    [
        ([100, 80, 50], operator.truediv, 2, [50, 40, 25]),
        ([100, 80, 50], operator.sub, 2, [98, 78, 48]),
        (2, operator.sub, [100, 80, 50], [-98, -78, -48]),
        (2, operator.add, [100, 80, 50], [102, 82, 52]),
        ([100, 80, 50], operator.add, 2, [102, 82, 52]),
        (2, operator.mul, [100, 80, 50], [200, 160, 100]),
        ([100, 80, 50], operator.mul, 2, [200, 160, 100]),
    ],
)
@pytest.mark.parametrize("array_maker", [MakeArray, MakeFixedArray])
def testScalarOperations(left, op, right, expected, array_maker) -> None:
    if isinstance(left, list):
        left = array_maker(left)
    if isinstance(right, list):
        right = array_maker(right)

    result = op(left, right)
    assert result == array_maker(expected)
